@@ -5,32 +5,14 @@
 /* initialize an fmtext with no data */
 void owl_fmtext_init_null(owl_fmtext *f)
 {
-  f->textlen = 0;
-  f->bufflen = 5;
-  f->textbuff = owl_malloc(5);
-  f->textbuff[0] = 0;
-  f->default_attrs = OWL_FMTEXT_ATTR_NONE;
-  f->default_fgcolor = OWL_COLOR_DEFAULT;
-  f->default_bgcolor = OWL_COLOR_DEFAULT;
+  f->buff = g_string_new("");
 }
 
 /* Clear the data from an fmtext, but don't deallocate memory. This
    fmtext can then be appended to again. */
 void owl_fmtext_clear(owl_fmtext *f)
 {
-  f->textlen = 0;
-  f->textbuff[0] = 0;
-  f->default_attrs = OWL_FMTEXT_ATTR_NONE;
-  f->default_fgcolor = OWL_COLOR_DEFAULT;
-  f->default_bgcolor = OWL_COLOR_DEFAULT;
-}
-
-static void _owl_fmtext_realloc(owl_fmtext *f, int newlen)
-{
-    if(newlen + 1 > f->bufflen) {
-      f->textbuff = owl_realloc(f->textbuff, newlen + 1);
-      f->bufflen = newlen+1;
-  }
+  g_string_truncate(f->buff, 0);
 }
 
 int owl_fmtext_is_format_char(gunichar c)
@@ -44,35 +26,26 @@ int owl_fmtext_is_format_char(gunichar c)
  */
 void owl_fmtext_append_attr(owl_fmtext *f, const char *text, char attr, short fgcolor, short bgcolor)
 {
-  char attrbuff[6];
-  int newlen, a = 0, fg = 0, bg = 0;
+  int a = 0, fg = 0, bg = 0;
   
   if (attr != OWL_FMTEXT_ATTR_NONE) a=1;
   if (fgcolor != OWL_COLOR_DEFAULT) fg=1;
   if (bgcolor != OWL_COLOR_DEFAULT) bg=1;
 
-  /* Plane-16 characters in UTF-8 are 4 bytes long. */
-  newlen = strlen(f->textbuff) + strlen(text) + (8 * (a + fg + bg));
-  _owl_fmtext_realloc(f, newlen);
-
   /* Set attributes */
   if (a)
-    strncat(f->textbuff, attrbuff,
-	    g_unichar_to_utf8(OWL_FMTEXT_UC_ATTR | attr, attrbuff));
+    g_string_append_unichar(f->buff, OWL_FMTEXT_UC_ATTR | attr);
   if (fg)
-    strncat(f->textbuff, attrbuff,
-	    g_unichar_to_utf8(OWL_FMTEXT_UC_FGCOLOR | fgcolor, attrbuff));
+    g_string_append_unichar(f->buff, OWL_FMTEXT_UC_FGCOLOR | fgcolor);
   if (bg)
-    strncat(f->textbuff, attrbuff,
-	    g_unichar_to_utf8(OWL_FMTEXT_UC_BGCOLOR | bgcolor, attrbuff));
-  
-  strcat(f->textbuff, text);
+    g_string_append_unichar(f->buff, OWL_FMTEXT_UC_BGCOLOR | bgcolor);
+
+  g_string_append(f->buff, text);
 
   /* Reset attributes */
-  if (bg) strcat(f->textbuff, OWL_FMTEXT_UTF8_BGDEFAULT);
-  if (fg) strcat(f->textbuff, OWL_FMTEXT_UTF8_FGDEFAULT);
-  if (a)  strcat(f->textbuff, OWL_FMTEXT_UTF8_ATTR_NONE);
-  f->textlen=newlen;
+  if (bg) g_string_append_unichar(f->buff, OWL_FMTEXT_UC_BGDEFAULT);
+  if (fg) g_string_append_unichar(f->buff, OWL_FMTEXT_UC_FGDEFAULT);
+  if (a)  g_string_append_unichar(f->buff, OWL_FMTEXT_UC_ATTR | OWL_FMTEXT_UC_ATTR);
 }
 
 /* Append normal, uncolored text 'text' to 'f' */
@@ -82,7 +55,7 @@ void owl_fmtext_append_normal(owl_fmtext *f, const char *text)
 }
 
 /* Append normal, uncolored text specified by format string to 'f' */
-void owl_fmtext_appendf_normal(owl_fmtext *f, const char *fmt, ...)
+void G_GNUC_PRINTF(2, 3) owl_fmtext_appendf_normal(owl_fmtext *f, const char *fmt, ...)
 {
   va_list ap;
   char *buff;
@@ -93,7 +66,7 @@ void owl_fmtext_appendf_normal(owl_fmtext *f, const char *fmt, ...)
   if (!buff)
     return;
   owl_fmtext_append_attr(f, buff, OWL_FMTEXT_ATTR_NONE, OWL_COLOR_DEFAULT, OWL_COLOR_DEFAULT);
-  owl_free(buff);
+  g_free(buff);
 }
 
 /* Append normal text 'text' to 'f' with color 'color' */
@@ -120,29 +93,6 @@ void owl_fmtext_append_reversebold(owl_fmtext *f, const char *text)
   owl_fmtext_append_attr(f, text, OWL_FMTEXT_ATTR_REVERSE | OWL_FMTEXT_ATTR_BOLD, OWL_COLOR_DEFAULT, OWL_COLOR_DEFAULT);
 }
 
-/* Add the attribute 'attr' to the default atts for the text in 'f' */
-void owl_fmtext_addattr(owl_fmtext *f, char attr)
-{
-  /* add the attribute to all text */
-  f->default_attrs |= attr;
-}
-
-/* Set the default foreground color for this fmtext to 'color'.
- * Only affects text that is colored default.
- */
-void owl_fmtext_colorize(owl_fmtext *f, int color)
-{
-  f->default_fgcolor = color;
-}
-
-/* Set the default foreground color for this fmtext to 'color'.
- * Only affects text that is colored default.
- */
-void owl_fmtext_colorizebg(owl_fmtext *f, int color)
-{
-  f->default_bgcolor = color;
-}
-
 /* Internal function. Parse attrbute character. */
 static void _owl_fmtext_update_attributes(gunichar c, char *attr, short *fgcolor, short *bgcolor)
 {
@@ -167,8 +117,8 @@ static void _owl_fmtext_update_attributes(gunichar c, char *attr, short *fgcolor
 static void _owl_fmtext_scan_attributes(const owl_fmtext *f, int start, char *attr, short *fgcolor, short *bgcolor)
 {
   const char *p;
-  p = strchr(f->textbuff, OWL_FMTEXT_UC_STARTBYTE_UTF8);
-  while (p && p < f->textbuff + start) {
+  p = strchr(f->buff->str, OWL_FMTEXT_UC_STARTBYTE_UTF8);
+  while (p && p < f->buff->str + start) {
     _owl_fmtext_update_attributes(g_utf8_get_char(p), attr, fgcolor, bgcolor);
     p = strchr(p+1, OWL_FMTEXT_UC_STARTBYTE_UTF8);
   }
@@ -180,47 +130,31 @@ static void _owl_fmtext_scan_attributes(const owl_fmtext *f, int start, char *at
  */
 static void _owl_fmtext_append_fmtext(owl_fmtext *f, const owl_fmtext *in, int start, int stop)
 {
-  char attrbuff[6];
-  int newlen, a = 0, fg = 0, bg = 0;
   char attr = 0;
   short fgcolor = OWL_COLOR_DEFAULT;
   short bgcolor = OWL_COLOR_DEFAULT;
 
   _owl_fmtext_scan_attributes(in, start, &attr, &fgcolor, &bgcolor);
-  if (attr != OWL_FMTEXT_ATTR_NONE) a=1;
-  if (fgcolor != OWL_COLOR_DEFAULT) fg=1;
-  if (bgcolor != OWL_COLOR_DEFAULT) bg=1;
 
-  /* We will reset to defaults after appending the text. We may need
-     to set initial attributes. */
-  newlen=strlen(f->textbuff)+(stop-start) + (4 * (a + fg + bg)) + 12;
-  _owl_fmtext_realloc(f, newlen);
+  if (attr != OWL_FMTEXT_ATTR_NONE)
+    g_string_append_unichar(f->buff, OWL_FMTEXT_UC_ATTR | attr);
+  if (fgcolor != OWL_COLOR_DEFAULT)
+    g_string_append_unichar(f->buff, OWL_FMTEXT_UC_FGCOLOR | fgcolor);
+  if (bgcolor != OWL_COLOR_DEFAULT)
+    g_string_append_unichar(f->buff, OWL_FMTEXT_UC_BGCOLOR | bgcolor);
 
-  if (a)
-    strncat(f->textbuff, attrbuff,
-	    g_unichar_to_utf8(OWL_FMTEXT_UC_ATTR | attr, attrbuff));
-  if (fg)
-    strncat(f->textbuff, attrbuff,
-	    g_unichar_to_utf8(OWL_FMTEXT_UC_FGCOLOR | fgcolor, attrbuff));
-  if (bg)
-    strncat(f->textbuff, attrbuff,
-	    g_unichar_to_utf8(OWL_FMTEXT_UC_BGCOLOR | bgcolor, attrbuff));
-
-  strncat(f->textbuff, in->textbuff+start, stop-start);
+  g_string_append_len(f->buff, in->buff->str+start, stop-start);
 
   /* Reset attributes */
-  strcat(f->textbuff, OWL_FMTEXT_UTF8_BGDEFAULT);
-  strcat(f->textbuff, OWL_FMTEXT_UTF8_FGDEFAULT);
-  strcat(f->textbuff, OWL_FMTEXT_UTF8_ATTR_NONE);
-
-  f->textbuff[newlen]='\0';
-  f->textlen=newlen;
+  g_string_append_unichar(f->buff, OWL_FMTEXT_UC_BGDEFAULT);
+  g_string_append_unichar(f->buff, OWL_FMTEXT_UC_FGDEFAULT);
+  g_string_append_unichar(f->buff, OWL_FMTEXT_UC_ATTR | OWL_FMTEXT_UC_ATTR);
 }
 
 /* append fmtext 'in' to 'f' */
 void owl_fmtext_append_fmtext(owl_fmtext *f, const owl_fmtext *in)
 {
-  _owl_fmtext_append_fmtext(f, in, 0, in->textlen);
+  _owl_fmtext_append_fmtext(f, in, 0, in->buff->len);
 
 }
 
@@ -236,9 +170,9 @@ void owl_fmtext_append_spaces(owl_fmtext *f, int nspaces)
 /* Return a plain version of the fmtext.  Caller is responsible for
  * freeing the return
  */
-char *owl_fmtext_print_plain(const owl_fmtext *f)
+CALLER_OWN char *owl_fmtext_print_plain(const owl_fmtext *f)
 {
-  return owl_strip_format_chars(f->textbuff);
+  return owl_strip_format_chars(f->buff->str);
 }
 
 static void _owl_fmtext_wattrset(WINDOW *w, int attrs)
@@ -249,16 +183,9 @@ static void _owl_fmtext_wattrset(WINDOW *w, int attrs)
   if (attrs & OWL_FMTEXT_ATTR_UNDERLINE) wattron(w, A_UNDERLINE);
 }
 
-static void _owl_fmtext_update_colorpair(short fg, short bg, short *pair)
-{
-  if (owl_global_get_hascolors(&g)) {
-    *pair = owl_fmtext_get_colorpair(fg, bg);
-  }
-}
-
 static void _owl_fmtext_wcolor_set(WINDOW *w, short pair)
 {
-  if (owl_global_get_hascolors(&g)) {
+  if (has_colors()) {
       wcolor_set(w,pair,NULL);
       wbkgdset(w, COLOR_PAIR(pair));
   }
@@ -267,7 +194,7 @@ static void _owl_fmtext_wcolor_set(WINDOW *w, short pair)
 /* add the formatted text to the curses window 'w'.  The window 'w'
  * must already be initiatlized with curses
  */
-static void _owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, int do_search)
+static void _owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, int do_search, char default_attrs, short default_fgcolor, short default_bgcolor)
 {
   /* char *tmpbuff; */
   /* int position, trans1, trans2, trans3, len, lastsame; */
@@ -280,13 +207,13 @@ static void _owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, int do_sear
     return;
   }
 
-  s = f->textbuff;
+  s = f->buff->str;
   /* Set default attributes. */
-  attr = f->default_attrs;
-  fg = f->default_fgcolor;
-  bg = f->default_bgcolor;
+  attr = default_attrs;
+  fg = default_fgcolor;
+  bg = default_bgcolor;
   _owl_fmtext_wattrset(w, attr);
-  _owl_fmtext_update_colorpair(fg, bg, &pair);
+  pair = owl_fmtext_get_colorpair(fg, bg);
   _owl_fmtext_wcolor_set(w, pair);
 
   /* Find next possible format character. */
@@ -300,7 +227,6 @@ static void _owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, int do_sear
       p[0] = '\0';
       if (do_search && owl_global_is_search_active(&g)) {
 	/* Search is active, so highlight search results. */
-	char tmp2;
 	int start, end;
 	while (owl_regex_compare(owl_global_get_search_re(&g), s, &start, &end) == 0) {
 	  /* Prevent an infinite loop matching the empty string. */
@@ -309,18 +235,12 @@ static void _owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, int do_sear
 
 	  /* Found search string, highlight it. */
 
-	  tmp2 = s[start];
-	  s[start] = '\0';
-	  waddstr(w, s);
-	  s[start] = tmp2;
+	  waddnstr(w, s, start);
 
 	  _owl_fmtext_wattrset(w, attr ^ OWL_FMTEXT_ATTR_REVERSE);
 	  _owl_fmtext_wcolor_set(w, pair);
 	  
-	  tmp2 = s[end];
-	  s[end] = '\0';
-	  waddstr(w, s + start);
-	  s[end] = tmp2;
+	  waddnstr(w, s + start, end - start);
 
 	  _owl_fmtext_wattrset(w, attr);
 	  _owl_fmtext_wcolor_set(w, pair);
@@ -332,19 +252,17 @@ static void _owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, int do_sear
       waddstr(w, s);
       p[0] = tmp;
 
-      /* Deal with new attributes. Initialize to defaults, then
-	 process all consecutive formatting characters. */
-      attr = f->default_attrs;
-      fg = f->default_fgcolor;
-      bg = f->default_bgcolor;
+      /* Deal with new attributes. Process all consecutive formatting
+       * characters, and then apply defaults where relevant. */
       while (owl_fmtext_is_format_char(g_utf8_get_char(p))) {
 	_owl_fmtext_update_attributes(g_utf8_get_char(p), &attr, &fg, &bg);
 	p = g_utf8_next_char(p);
       }
-      _owl_fmtext_wattrset(w, attr | f->default_attrs);
-      if (fg == OWL_COLOR_DEFAULT) fg = f->default_fgcolor;
-      if (bg == OWL_COLOR_DEFAULT) bg = f->default_bgcolor;
-      _owl_fmtext_update_colorpair(fg, bg, &pair);
+      attr |= default_attrs;
+      if (fg == OWL_COLOR_DEFAULT) fg = default_fgcolor;
+      if (bg == OWL_COLOR_DEFAULT) bg = default_bgcolor;
+      _owl_fmtext_wattrset(w, attr);
+      pair = owl_fmtext_get_colorpair(fg, bg);
       _owl_fmtext_wcolor_set(w, pair);
 
       /* Advance to next non-formatting character. */
@@ -361,14 +279,14 @@ static void _owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, int do_sear
   wbkgdset(w, 0);
 }
 
-void owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w)
+void owl_fmtext_curs_waddstr(const owl_fmtext *f, WINDOW *w, char default_attrs, short default_fgcolor, short default_bgcolor)
 {
-  _owl_fmtext_curs_waddstr(f, w, 1);
+  _owl_fmtext_curs_waddstr(f, w, 1, default_attrs, default_fgcolor, default_bgcolor);
 }
 
-void owl_fmtext_curs_waddstr_without_search(const owl_fmtext *f, WINDOW *w)
+void owl_fmtext_curs_waddstr_without_search(const owl_fmtext *f, WINDOW *w, char default_attrs, short default_fgcolor, short default_bgcolor)
 {
-  _owl_fmtext_curs_waddstr(f, w, 0);
+  _owl_fmtext_curs_waddstr(f, w, 0, default_attrs, default_fgcolor, default_bgcolor);
 }
 
 /* Expands tabs. Tabs are expanded as if given an initial indent of start. */
@@ -376,24 +294,19 @@ void owl_fmtext_expand_tabs(const owl_fmtext *in, owl_fmtext *out, int start) {
   int col = start, numcopied = 0;
   char *ptr;
 
-  /* Copy the default attributes. */
-  out->default_attrs = in->default_attrs;
-  out->default_fgcolor = in->default_fgcolor;
-  out->default_bgcolor = in->default_bgcolor;
-
-  for (ptr = in->textbuff;
-       ptr < in->textbuff + in->textlen;
+  for (ptr = in->buff->str;
+       ptr < in->buff->str + in->buff->len;
        ptr = g_utf8_next_char(ptr)) {
     gunichar c = g_utf8_get_char(ptr);
     int chwidth;
     if (c == '\t') {
       /* Copy up to this tab */
-      _owl_fmtext_append_fmtext(out, in, numcopied, ptr - in->textbuff);
+      _owl_fmtext_append_fmtext(out, in, numcopied, ptr - in->buff->str);
       /* and then copy spaces for the tab. */
       chwidth = OWL_TAB_WIDTH - (col % OWL_TAB_WIDTH);
       owl_fmtext_append_spaces(out, chwidth);
       col += chwidth;
-      numcopied = g_utf8_next_char(ptr) - in->textbuff;
+      numcopied = g_utf8_next_char(ptr) - in->buff->str;
     } else {
       /* Just update col. We'll append later. */
       if (c == '\n') {
@@ -404,8 +317,8 @@ void owl_fmtext_expand_tabs(const owl_fmtext *in, owl_fmtext *out, int start) {
     }
   }
   /* Append anything we've missed. */
-  if (numcopied < in->textlen)
-    _owl_fmtext_append_fmtext(out, in, numcopied, in->textlen);
+  if (numcopied < in->buff->len)
+    _owl_fmtext_append_fmtext(out, in, numcopied, in->buff->len);
 }
 
 /* start with line 'aline' (where the first line is 0) and print
@@ -417,7 +330,7 @@ int owl_fmtext_truncate_lines(const owl_fmtext *in, int aline, int lines, owl_fm
   int i, offset;
   
   /* find the starting line */
-  ptr1 = in->textbuff;
+  ptr1 = in->buff->str;
   for (i = 0; i < aline; i++) {
     ptr1 = strchr(ptr1, '\n');
     if (!ptr1) return(-1);
@@ -426,20 +339,15 @@ int owl_fmtext_truncate_lines(const owl_fmtext *in, int aline, int lines, owl_fm
   
   /* ptr1 now holds the starting point */
 
-  /* copy the default attributes */
-  out->default_attrs = in->default_attrs;
-  out->default_fgcolor = in->default_fgcolor;
-  out->default_bgcolor = in->default_bgcolor;
-    
   /* copy in the next 'lines' lines */
   if (lines < 1) return(-1);
 
   for (i = 0; i < lines; i++) {
-    offset = ptr1 - in->textbuff;
+    offset = ptr1 - in->buff->str;
     ptr2 = strchr(ptr1, '\n');
     if (!ptr2) {
       /* Copy to the end of the buffer. */
-      _owl_fmtext_append_fmtext(out, in, offset, in->textlen);
+      _owl_fmtext_append_fmtext(out, in, offset, in->buff->len);
       return(-1);
     }
     /* Copy up to, and including, the new line. */
@@ -455,13 +363,8 @@ void _owl_fmtext_truncate_cols_internal(const owl_fmtext *in, int acol, int bcol
   const char *ptr_s, *ptr_e, *ptr_c, *last;
   int col, st, padding, chwidth;
 
-  /* copy the default attributes */
-  out->default_attrs = in->default_attrs;
-  out->default_fgcolor = in->default_fgcolor;
-  out->default_bgcolor = in->default_bgcolor;
-
-  last=in->textbuff+in->textlen-1;
-  ptr_s=in->textbuff;
+  last = in->buff->str + in->buff->len - 1;
+  ptr_s = in->buff->str;
   while (ptr_s <= last) {
     ptr_e=strchr(ptr_s, '\n');
     if (!ptr_e) {
@@ -503,18 +406,18 @@ void _owl_fmtext_truncate_cols_internal(const owl_fmtext *in, int acol, int bcol
       owl_fmtext_append_spaces(out, padding);
       if (ptr_c == ptr_e) {
 	/* We made it to the newline. Append up to, and including it. */
-	_owl_fmtext_append_fmtext(out, in, ptr_s - in->textbuff, ptr_c - in->textbuff + 1);
+	_owl_fmtext_append_fmtext(out, in, ptr_s - in->buff->str, ptr_c - in->buff->str + 1);
       }
       else if (chwidth > 1) {
         /* Last char is wide, truncate. */
-        _owl_fmtext_append_fmtext(out, in, ptr_s - in->textbuff, ptr_c - in->textbuff);
+        _owl_fmtext_append_fmtext(out, in, ptr_s - in->buff->str, ptr_c - in->buff->str);
         owl_fmtext_append_normal(out, "\n");
       }
       else {
         /* Last char fits perfectly, We stop at the next char to make
 	 * sure we get it all. */
         ptr_c = g_utf8_next_char(ptr_c);
-        _owl_fmtext_append_fmtext(out, in, ptr_s - in->textbuff, ptr_c - in->textbuff);
+        _owl_fmtext_append_fmtext(out, in, ptr_s - in->buff->str, ptr_c - in->buff->str);
       }
     }
     else {
@@ -541,7 +444,7 @@ void owl_fmtext_truncate_cols(const owl_fmtext *in, int acol, int bcol, owl_fmte
   owl_fmtext notabs;
 
   /* _owl_fmtext_truncate_cols_internal cannot handle tabs. */
-  if (strchr(in->textbuff, '\t')) {
+  if (strchr(in->buff->str, '\t')) {
     owl_fmtext_init_null(&notabs);
     owl_fmtext_expand_tabs(in, &notabs, 0);
     _owl_fmtext_truncate_cols_internal(&notabs, acol, bcol, out);
@@ -558,17 +461,17 @@ int owl_fmtext_num_lines(const owl_fmtext *f)
   char *lastbreak, *p;
 
   lines=0;
-  lastbreak = f->textbuff;
-  for (i=0; i<f->textlen; i++) {
-    if (f->textbuff[i]=='\n') {
-      lastbreak = f->textbuff + i;
+  lastbreak = f->buff->str;
+  for (i = 0; i < f->buff->len; i++) {
+    if (f->buff->str[i]=='\n') {
+      lastbreak = f->buff->str + i;
       lines++;
     }
   }
 
   /* Check if there's a trailing line; formatting characters don't count. */
   for (p = g_utf8_next_char(lastbreak);
-       p < f->textbuff + f->textlen;
+       p < f->buff->str + f->buff->len;
        p = g_utf8_next_char(p)) {
     if (!owl_fmtext_is_format_char(g_utf8_get_char(p))) {
       lines++;
@@ -587,10 +490,10 @@ int owl_fmtext_num_lines(const owl_fmtext *f)
 int owl_fmtext_line_number(const owl_fmtext *f, int offset)
 {
   int i, lineno = 0;
-  if (offset >= f->textlen)
-    offset = f->textlen - 1;
+  if (offset >= f->buff->len)
+    offset = f->buff->len - 1;
   for (i = 0; i < offset; i++) {
-    if (f->textbuff[i] == '\n')
+    if (f->buff->str[i] == '\n')
       lineno++;
   }
   return lineno;
@@ -602,55 +505,31 @@ void owl_fmtext_line_extents(const owl_fmtext *f, int lineno, int *o_start, int
 {
   int start, end;
   char *newline;
-  for (start = 0; lineno > 0 && start < f->textlen; start++) {
-    if (f->textbuff[start] == '\n')
+  for (start = 0; lineno > 0 && start < f->buff->len; start++) {
+    if (f->buff->str[start] == '\n')
       lineno--;
   }
-  newline = strchr(f->textbuff + start, '\n');
+  newline = strchr(f->buff->str + start, '\n');
   /* Include the newline, if it is there. */
-  end = newline ? newline - f->textbuff + 1 : f->textlen;
+  end = newline ? newline - f->buff->str + 1 : f->buff->len;
   if (o_start) *o_start = start;
   if (o_end) *o_end = end;
 }
 
 const char *owl_fmtext_get_text(const owl_fmtext *f)
 {
-  return(f->textbuff);
+  return f->buff->str;
 }
 
 int owl_fmtext_num_bytes(const owl_fmtext *f)
 {
-  return(f->textlen);
-}
-
-/* set the charater at 'index' to be 'char'.  If index is out of
- * bounds don't do anything. If c or char at index is not ASCII, don't
- * do anything because it's not UTF-8 safe. */
-void owl_fmtext_set_char(owl_fmtext *f, int index, char ch)
-{
-  if ((index < 0) || (index > f->textlen-1)) return;
-  /* NOT ASCII*/
-  if (f->textbuff[index] & 0x80 || ch & 0x80) return; 
-  f->textbuff[index]=ch;
+  return f->buff->len;
 }
 
 /* Make a copy of the fmtext 'src' into 'dst' */
 void owl_fmtext_copy(owl_fmtext *dst, const owl_fmtext *src)
 {
-  int mallocsize;
-
-  if (src->textlen==0) {
-    mallocsize=5;
-  } else {
-    mallocsize=src->textlen+2;
-  }
-  dst->textlen=src->textlen;
-  dst->bufflen=mallocsize;
-  dst->textbuff=owl_malloc(mallocsize);
-  memcpy(dst->textbuff, src->textbuff, src->textlen+1);
-  dst->default_attrs = src->default_attrs;
-  dst->default_fgcolor = src->default_fgcolor;
-  dst->default_bgcolor = src->default_bgcolor;
+  dst->buff = g_string_new(src->buff->str);
 }
 
 /* Search 'f' for the regex 're' for matches starting at
@@ -660,8 +539,8 @@ void owl_fmtext_copy(owl_fmtext *dst, const owl_fmtext *src)
 int owl_fmtext_search(const owl_fmtext *f, const owl_regex *re, int start)
 {
   int offset;
-  if (start > f->textlen ||
-      owl_regex_compare(re, f->textbuff + start, &offset, NULL) != 0)
+  if (start > f->buff->len ||
+      owl_regex_compare(re, f->buff->str + start, &offset, NULL) != 0)
     return -1;
   return offset + start;
 }
@@ -689,11 +568,11 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
       return;
     } else if (ptr[0]=='@') {
       /* add the text up to this point then deal with the stack */
-      buff=owl_malloc(ptr-txtptr+20);
+      buff=g_new(char, ptr-txtptr+20);
       strncpy(buff, txtptr, ptr-txtptr);
       buff[ptr-txtptr]='\0';
       owl_fmtext_append_attr(f, buff, curattrs, curcolor, OWL_COLOR_DEFAULT);
-      owl_free(buff);
+      g_free(buff);
 
       /* update pointer to point at the @ */
       txtptr=ptr;
@@ -724,7 +603,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
 
       /* check what command we've got, push it on the stack, start
          using it, and continue ... unless it's a color command */
-      buff=owl_malloc(tmpptr-ptr+20);
+      buff=g_new(char, tmpptr-ptr+20);
       strncpy(buff, ptr, tmpptr-ptr);
       buff[tmpptr-ptr]='\0';
       if (!strcasecmp(buff, "@bold")) {
@@ -734,7 +613,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
         stacksize++;
         curattrs|=OWL_FMTEXT_ATTR_BOLD;
         txtptr+=6;
-        owl_free(buff);
+        g_free(buff);
         continue;
       } else if (!strcasecmp(buff, "@b")) {
         attrstack[stacksize]=OWL_FMTEXT_ATTR_BOLD;
@@ -743,7 +622,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
         stacksize++;
         curattrs|=OWL_FMTEXT_ATTR_BOLD;
         txtptr+=3;
-        owl_free(buff);
+        g_free(buff);
         continue;
       } else if (!strcasecmp(buff, "@i")) {
         attrstack[stacksize]=OWL_FMTEXT_ATTR_UNDERLINE;
@@ -752,7 +631,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
         stacksize++;
         curattrs|=OWL_FMTEXT_ATTR_UNDERLINE;
         txtptr+=3;
-        owl_free(buff);
+        g_free(buff);
         continue;
       } else if (!strcasecmp(buff, "@italic")) {
         attrstack[stacksize]=OWL_FMTEXT_ATTR_UNDERLINE;
@@ -761,7 +640,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
         stacksize++;
         curattrs|=OWL_FMTEXT_ATTR_UNDERLINE;
         txtptr+=8;
-        owl_free(buff);
+        g_free(buff);
         continue;
       } else if (!strcasecmp(buff, "@")) {
 	attrstack[stacksize]=OWL_FMTEXT_ATTR_NONE;
@@ -769,15 +648,14 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
 	colorstack[stacksize]=curcolor;
         stacksize++;
         txtptr+=2;
-        owl_free(buff);
+        g_free(buff);
         continue;
 
         /* if it's a color read the color, set the current color and
            continue */
       } else if (!strcasecmp(buff, "@color") 
-                 && owl_global_get_hascolors(&g)
                  && owl_global_is_colorztext(&g)) {
-        owl_free(buff);
+        g_free(buff);
         txtptr+=7;
         tmpptr=strpbrk(txtptr, "@{[<()>]}");
         if (tmpptr &&
@@ -787,7 +665,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
              (txtptr[-1]=='{' && tmpptr[0]=='}'))) {
 
           /* grab the color name */
-          buff=owl_malloc(tmpptr-txtptr+20);
+          buff=g_new(char, tmpptr-txtptr+20);
           strncpy(buff, txtptr, tmpptr-txtptr);
           buff[tmpptr-txtptr]='\0';
 
@@ -795,7 +673,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
           curcolor=owl_util_string_to_color(buff);
           if (curcolor == OWL_COLOR_INVALID)
 	      curcolor = OWL_COLOR_DEFAULT;
-          owl_free(buff);
+          g_free(buff);
           txtptr=tmpptr+1;
           continue;
 
@@ -807,6 +685,7 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
         /* if we didn't understand it, we'll print it.  This is different from zwgc
          * but zwgc seems to be smarter about some screw cases than I am
          */
+        g_free(buff);
         owl_fmtext_append_attr(f, "@", curattrs, curcolor, OWL_COLOR_DEFAULT);
         txtptr++;
         continue;
@@ -814,22 +693,22 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
 
     } else if (ptr[0]=='}' || ptr[0]==']' || ptr[0]==')' || ptr[0]=='>') {
       /* add the text up to this point first */
-      buff=owl_malloc(ptr-txtptr+20);
+      buff=g_new(char, ptr-txtptr+20);
       strncpy(buff, txtptr, ptr-txtptr);
       buff[ptr-txtptr]='\0';
       owl_fmtext_append_attr(f, buff, curattrs, curcolor, OWL_COLOR_DEFAULT);
-      owl_free(buff);
+      g_free(buff);
 
       /* now deal with the closer */
       txtptr=ptr;
 
       /* first, if the stack is empty we must bail (just print and go) */
       if (stacksize==0) {
-        buff=owl_malloc(5);
+        buff=g_new(char, 5);
         buff[0]=ptr[0];
         buff[1]='\0';
         owl_fmtext_append_attr(f, buff, curattrs, curcolor, OWL_COLOR_DEFAULT);
-        owl_free(buff);
+        g_free(buff);
         txtptr++;
         continue;
       }
@@ -851,21 +730,21 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
         continue;
       } else {
         /* otherwise print and continue */
-        buff=owl_malloc(5);
+        buff=g_new(char, 5);
         buff[0]=ptr[0];
         buff[1]='\0';
         owl_fmtext_append_attr(f, buff, curattrs, curcolor, OWL_COLOR_DEFAULT);
-        owl_free(buff);
+        g_free(buff);
         txtptr++;
         continue;
       }
     } else {
       /* we've found an unattached opener, print everything and move on */
-      buff=owl_malloc(ptr-txtptr+20);
+      buff=g_new(char, ptr-txtptr+20);
       strncpy(buff, txtptr, ptr-txtptr+1);
       buff[ptr-txtptr+1]='\0';
       owl_fmtext_append_attr(f, buff, curattrs, curcolor, OWL_COLOR_DEFAULT);
-      owl_free(buff);
+      g_free(buff);
       txtptr=ptr+1;
       continue;
     }
@@ -876,25 +755,24 @@ void owl_fmtext_append_ztext(owl_fmtext *f, const char *text)
  * joins the elements together with join_with. 
  * If format_fn is specified, passes it the list element value
  * and it will return a string which this needs to free. */
-void owl_fmtext_append_list(owl_fmtext *f, const owl_list *l, const char *join_with, char *(format_fn)(const char *))
+void owl_fmtext_append_list(owl_fmtext *f, const GPtrArray *l, const char *join_with, char *(format_fn)(const char *))
 {
-  int i, size;
+  int i;
   const char *elem;
   char *text;
 
-  size = owl_list_get_size(l);
-  for (i=0; i<size; i++) {
-    elem = owl_list_get_element(l,i);
+  for (i = 0; i < l->len; i++) {
+    elem = l->pdata[i];
     if (elem && format_fn) {
       text = format_fn(elem);
       if (text) {
         owl_fmtext_append_normal(f, text);
-        owl_free(text);
+        g_free(text);
       }
     } else if (elem) {
       owl_fmtext_append_normal(f, elem);
     }
-    if ((i < size-1) && join_with) {
+    if ((i < l->len - 1) && join_with) {
       owl_fmtext_append_normal(f, join_with);
     }
   }
@@ -903,7 +781,8 @@ void owl_fmtext_append_list(owl_fmtext *f, const owl_list *l, const char *join_w
 /* Free all memory allocated by the object */
 void owl_fmtext_cleanup(owl_fmtext *f)
 {
-  if (f->textbuff) owl_free(f->textbuff);
+  if (f->buff) g_string_free(f->buff, true);
+  f->buff = NULL;
 }
 
 /*** Color Pair manager ***/
@@ -912,9 +791,9 @@ void owl_fmtext_init_colorpair_mgr(owl_colorpair_mgr *cpmgr)
   /* This could be a bitarray if we wanted to save memory. */
   short i;
   /* The test is <= because we allocate COLORS+1 entries. */
-  cpmgr->pairs = owl_malloc((COLORS+1) * sizeof(short*));
+  cpmgr->pairs = g_new(short *, COLORS + 1);
   for(i = 0; i <= COLORS; i++) {
-    cpmgr->pairs[i] = owl_malloc((COLORS+1) * sizeof(short));
+    cpmgr->pairs[i] = g_new(short, COLORS + 1);
   }
   owl_fmtext_reset_colorpairs(cpmgr);
 }
@@ -932,7 +811,7 @@ void owl_fmtext_reset_colorpairs(owl_colorpair_mgr *cpmgr)
       cpmgr->pairs[i][j] = -1;
     }
   }
-  if (owl_global_get_hascolors(&g)) {
+  if (has_colors()) {
     for(i = 0; i < 8; i++) {
       short fg, bg;
       if (i >= COLORS) continue;
@@ -948,6 +827,9 @@ short owl_fmtext_get_colorpair(int fg, int bg)
   owl_colorpair_mgr *cpmgr;
   short pair;
 
+  if (!has_colors())
+    return 0;
+
   /* Sanity (Bounds) Check */
   if (fg > COLORS || fg < OWL_COLOR_DEFAULT) fg = OWL_COLOR_DEFAULT;
   if (bg > COLORS || bg < OWL_COLOR_DEFAULT) bg = OWL_COLOR_DEFAULT;
@@ -964,7 +846,7 @@ short owl_fmtext_get_colorpair(int fg, int bg)
   pair = cpmgr->pairs[fg+1][bg+1];
   if (!(pair != -1 && pair < cpmgr->next)) {
     /* If we didn't find a pair, search for a free one to assign. */
-    pair = (cpmgr->next < COLOR_PAIRS) ? cpmgr->next : -1;
+    pair = (cpmgr->next < owl_util_get_colorpairs()) ? cpmgr->next : -1;
     if (pair != -1) {
       /* We found a free pair, initialize it. */
       init_pair(pair, fg, bg);
