@@ -8,26 +8,16 @@
 #include <time.h>
 #include "owl.h"
 
-#ifndef MAXHOSTNAMELEN
-#define MAXHOSTNAMELEN 256
-#endif
-
 static void _owl_global_init_windows(owl_global *g);
 
 void owl_global_init(owl_global *g) {
-  struct hostent *hent;
-  char hostname[MAXHOSTNAMELEN];
   char *cd;
+  const char *homedir;
 
   g_type_init();
+  g_thread_init(NULL);
 
-  gethostname(hostname, MAXHOSTNAMELEN);
-  hent=gethostbyname(hostname);
-  if (!hent) {
-    g->thishost=owl_strdup("localhost");
-  } else {
-    g->thishost=owl_strdup(hent->h_name);
-  }
+  owl_select_init();
 
   g->lines=LINES;
   g->cols=COLS;
@@ -56,17 +46,12 @@ void owl_global_init(owl_global *g) {
 
   owl_dict_create(&(g->filters));
   g->filterlist = NULL;
-  owl_list_create(&(g->puntlist));
+  g->puntlist = g_ptr_array_new();
   g->messagequeue = g_queue_new();
   owl_dict_create(&(g->styledict));
   g->curmsg_vert_offset=0;
   g->resizepending=0;
   g->direction=OWL_DIRECTION_DOWNWARDS;
-  g->zaway=0;
-  if (has_colors()) {
-    g->hascolors=1;
-  }
-  g->colorpairs=COLOR_PAIRS;
   owl_fmtext_init_colorpair_mgr(&(g->cpmgr));
   g->debug=OWL_DEBUG;
   owl_regex_init(&g->search_re);
@@ -78,18 +63,22 @@ void owl_global_init(owl_global *g) {
   owl_global_set_no_have_config(g);
   owl_history_init(&(g->msghist));
   owl_history_init(&(g->cmdhist));
-  owl_history_set_norepeats(&(g->cmdhist));
   g->nextmsgid=0;
 
   /* Fill in some variables which don't have constant defaults */
-  /* TODO: come back later and check passwd file first */
-  g->homedir=owl_strdup(getenv("HOME"));
+
+  /* glib's g_get_home_dir prefers passwd entries to $HOME, so we
+   * explicitly check getenv first. */
+  homedir = getenv("HOME");
+  if (!homedir)
+    homedir = g_get_home_dir();
+  g->homedir = g_strdup(homedir);
 
   g->confdir = NULL;
   g->startupfile = NULL;
-  cd = owl_sprintf("%s/%s", g->homedir, OWL_CONFIG_DIR);
+  cd = g_build_filename(g->homedir, OWL_CONFIG_DIR, NULL);
   owl_global_set_confdir(g, cd);
-  owl_free(cd);
+  g_free(cd);
 
   owl_messagelist_create(&(g->msglist));
 
@@ -106,7 +95,6 @@ void owl_global_init(owl_global *g) {
   owl_global_set_no_doaimevents(g);
 
   owl_errqueue_init(&(g->errqueue));
-  g->got_err_signal=0;
 
   owl_zbuddylist_create(&(g->zbuddies));
 
@@ -114,10 +102,10 @@ void owl_global_init(owl_global *g) {
   g->pseudologin_notify = 0;
 
   owl_message_init_fmtext_cache();
-  owl_list_create(&(g->io_dispatch_list));
-  owl_list_create(&(g->psa_list));
-  g->timerlist = NULL;
-  g->interrupted = FALSE;
+  g->kill_buffer = NULL;
+
+  g->interrupt_count = 0;
+  g->interrupt_lock = g_mutex_new();
 }
 
 static void _owl_global_init_windows(owl_global *g)
@@ -182,7 +170,8 @@ void owl_global_push_context_obj(owl_global *g, owl_context *c)
 
 /* Pops the current context from the context stack and returns it. Caller is
  * responsible for freeing. */
-owl_context *owl_global_pop_context_no_delete(owl_global *g) {
+CALLER_OWN owl_context *owl_global_pop_context_no_delete(owl_global *g)
+{
   owl_context *c;
   if (!g->context_stack)
     return NULL;
@@ -354,7 +343,7 @@ void owl_global_set_typwin_inactive(owl_global *g) {
 /* resize */
 
 void owl_global_set_resize_pending(owl_global *g) {
-  g->resizepending=1;
+  g->resizepending = true;
 }
 
 const char *owl_global_get_homedir(const owl_global *g) {
@@ -371,10 +360,10 @@ const char *owl_global_get_confdir(const owl_global *g) {
  * Setting this also sets startupfile to confdir/startup
  */
 void owl_global_set_confdir(owl_global *g, const char *cd) {
-  owl_free(g->confdir);
-  g->confdir = owl_strdup(cd);
-  owl_free(g->startupfile);
-  g->startupfile = owl_sprintf("%s/startup", cd);
+  g_free(g->confdir);
+  g->confdir = g_strdup(cd);
+  g_free(g->startupfile);
+  g->startupfile = g_build_filename(cd, "startup", NULL);
 }
 
 const char *owl_global_get_startupfile(const owl_global *g) {
@@ -454,7 +443,7 @@ void owl_global_get_terminal_size(int *lines, int *cols) {
 void owl_global_check_resize(owl_global *g) {
   /* resize the screen.  If lines or cols is 0 use the terminal size */
   if (!g->resizepending) return;
-  g->resizepending = 0;
+  g->resizepending = false;
 
   owl_global_get_terminal_size(&g->lines, &g->cols);
   owl_window_resize(owl_window_get_screen(), g->lines, g->cols);
@@ -491,11 +480,6 @@ time_t owl_global_get_idletime(const owl_global *g) {
   return(time(NULL)-g->lastinputtime);
 }
 
-const char *owl_global_get_hostname(const owl_global *g) {
-  if (g->thishost) return(g->thishost);
-  return("");
-}
-
 /* viewwin */
 
 owl_viewwin *owl_global_get_viewwin(owl_global *g) {
@@ -519,22 +503,9 @@ void owl_global_set_curmsg_vert_offset(owl_global *g, int i) {
 
 /* startup args */
 
-void owl_global_set_startupargs(owl_global *g, int argc, const char *const *argv) {
-  int i, len;
-
-  if (g->startupargs) owl_free(g->startupargs);
-  
-  len=0;
-  for (i=0; i<argc; i++) {
-    len+=strlen(argv[i])+5;
-  }
-  g->startupargs=owl_malloc(len+5);
-
-  strcpy(g->startupargs, "");
-  for (i=0; i<argc; i++) {
-    sprintf(g->startupargs + strlen(g->startupargs), "%s ", argv[i]);
-  }
-  g->startupargs[strlen(g->startupargs)-1]='\0';
+void owl_global_set_startupargs(owl_global *g, int argc, char **argv) {
+  g_free(g->startupargs);
+  g->startupargs = g_strjoinv(" ", argv);
 }
 
 const char *owl_global_get_startupargs(const owl_global *g) {
@@ -569,11 +540,11 @@ static void owl_global_delete_filter_ent(void *data)
   owl_global_filter_ent *e = data;
   e->g->filterlist = g_list_remove(e->g->filterlist, e->f);
   owl_filter_delete(e->f);
-  owl_free(e);
+  g_free(e);
 }
 
 void owl_global_add_filter(owl_global *g, owl_filter *f) {
-  owl_global_filter_ent *e = owl_malloc(sizeof *e);
+  owl_global_filter_ent *e = g_new(owl_global_filter_ent, 1);
   e->g = g;
   e->f = f;
 
@@ -600,39 +571,25 @@ owl_view *owl_global_get_current_view(owl_global *g) {
   return(&(g->current_view));
 }
 
-/* has colors */
-
-int owl_global_get_hascolors(const owl_global *g) {
-  if (g->hascolors) return(1);
-  return(0);
-}
-
-/* color pairs */
-
-int owl_global_get_colorpairs(const owl_global *g) {
-  return(g->colorpairs);
-}
-
 owl_colorpair_mgr *owl_global_get_colorpair_mgr(owl_global *g) {
   return(&(g->cpmgr));
 }
 
 /* puntlist */
 
-owl_list *owl_global_get_puntlist(owl_global *g) {
-  return(&(g->puntlist));
+GPtrArray *owl_global_get_puntlist(owl_global *g) {
+  return g->puntlist;
 }
 
 int owl_global_message_is_puntable(owl_global *g, const owl_message *m) {
-  const owl_list *pl;
-  int i, j;
+  const GPtrArray *pl;
+  int i;
 
-  pl=owl_global_get_puntlist(g);
-  j=owl_list_get_size(pl);
-  for (i=0; i<j; i++) {
-    if (owl_filter_message_match(owl_list_get_element(pl, i), m)) return(1);
+  pl = owl_global_get_puntlist(g);
+  for (i = 0; i < pl->len; i++) {
+    if (owl_filter_message_match(pl->pdata[i], m)) return 1;
   }
-  return(0);
+  return 0;
 }
 
 int owl_global_should_followlast(owl_global *g) {
@@ -704,12 +661,12 @@ void owl_global_set_aimloggedin(owl_global *g, const char *screenname)
 {
   char *sn_escaped;
   g->aim_loggedin=1;
-  if (g->aim_screenname) owl_free(g->aim_screenname);
-  if (g->aim_screenname_for_filters) owl_free(g->aim_screenname_for_filters);
-  g->aim_screenname=owl_strdup(screenname);
+  if (g->aim_screenname) g_free(g->aim_screenname);
+  if (g->aim_screenname_for_filters) g_free(g->aim_screenname_for_filters);
+  g->aim_screenname=g_strdup(screenname);
   sn_escaped = owl_text_quote(screenname, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
   g->aim_screenname_for_filters = owl_arg_quote(sn_escaped);
-  owl_free(sn_escaped);
+  g_free(sn_escaped);
 }
 
 void owl_global_set_aimnologgedin(owl_global *g)
@@ -717,20 +674,26 @@ void owl_global_set_aimnologgedin(owl_global *g)
   g->aim_loggedin=0;
 }
 
-int owl_global_is_doaimevents(const owl_global *g)
+bool owl_global_is_doaimevents(const owl_global *g)
 {
-  if (g->aim_doprocessing) return(1);
-  return(0);
+  return g->aim_event_source != NULL;
 }
 
 void owl_global_set_doaimevents(owl_global *g)
 {
-  g->aim_doprocessing=1;
+  if (g->aim_event_source)
+    return;
+  g->aim_event_source = owl_aim_event_source_new(owl_global_get_aimsess(g));
+  g_source_attach(g->aim_event_source, NULL);
 }
 
 void owl_global_set_no_doaimevents(owl_global *g)
 {
-  g->aim_doprocessing=0;
+  if (!g->aim_event_source)
+    return;
+  g_source_destroy(g->aim_event_source);
+  g_source_unref(g->aim_event_source);
+  g->aim_event_source = NULL;
 }
 
 aim_session_t *owl_global_get_aimsess(owl_global *g)
@@ -759,7 +722,7 @@ void owl_global_messagequeue_addmsg(owl_global *g, owl_message *m)
  * is empty.  The caller should free the message after using it, if
  * necessary.
  */
-owl_message *owl_global_messagequeue_popmsg(owl_global *g)
+CALLER_OWN owl_message *owl_global_messagequeue_popmsg(owl_global *g)
 {
   owl_message *out;
 
@@ -788,10 +751,9 @@ const owl_style *owl_global_get_style_by_name(const owl_global *g, const char *n
   return owl_dict_find_element(&(g->styledict), name);
 }
 
-/* creates a list and fills it in with keys.  duplicates the keys, 
- * so they will need to be freed by the caller. */
-int owl_global_get_style_names(const owl_global *g, owl_list *l) {
-  return owl_dict_get_keys(&(g->styledict), l);
+CALLER_OWN GPtrArray *owl_global_get_style_names(const owl_global *g)
+{
+  return owl_dict_get_keys(&g->styledict);
 }
 
 void owl_global_add_style(owl_global *g, owl_style *s)
@@ -850,30 +812,6 @@ owl_errqueue *owl_global_get_errqueue(owl_global *g)
   return(&(g->errqueue));
 }
 
-void owl_global_set_errsignal(owl_global *g, int signum, siginfo_t *siginfo)
-{
-  g->got_err_signal = signum;
-  if (siginfo) {
-    g->err_signal_info = *siginfo;
-  } else {
-    siginfo_t si;
-    memset(&si, 0, sizeof(si));
-    g->err_signal_info = si;
-  }
-}
-
-int owl_global_get_errsignal_and_clear(owl_global *g, siginfo_t *siginfo)
-{
-  int signum;
-  if (siginfo && g->got_err_signal) {
-    *siginfo = g->err_signal_info;
-  } 
-  signum = g->got_err_signal;
-  g->got_err_signal = 0;
-  return signum;
-}
-
-
 owl_zbuddylist *owl_global_get_zephyr_buddylist(owl_global *g)
 {
   return(&(g->zbuddies));
@@ -899,33 +837,6 @@ struct termios *owl_global_get_startup_tio(owl_global *g)
   return(&(g->startup_tio));
 }
 
-owl_list *owl_global_get_io_dispatch_list(owl_global *g)
-{
-  return &(g->io_dispatch_list);
-}
-
-owl_list *owl_global_get_psa_list(owl_global *g)
-{
-  return &(g->psa_list);
-}
-
-GList **owl_global_get_timerlist(owl_global *g)
-{
-  return &(g->timerlist);
-}
-
-int owl_global_is_interrupted(const owl_global *g) {
-  return g->interrupted;
-}
-
-void owl_global_set_interrupted(owl_global *g) {
-  g->interrupted = 1;
-}
-
-void owl_global_unset_interrupted(owl_global *g) {
-  g->interrupted = 0;
-}
-
 void owl_global_setup_default_filters(owl_global *g)
 {
   int i;
@@ -970,15 +881,43 @@ FILE *owl_global_get_debug_file_handle(owl_global *g) {
 
     g->debug_file = NULL;
 
-    path = owl_sprintf("%s.%d", filename, getpid());
+    path = g_strdup_printf("%s.%d", filename, getpid());
     fd = open(path, O_CREAT|O_WRONLY|O_EXCL, 0600);
-    owl_free(path);
+    g_free(path);
 
     if (fd >= 0)
       g->debug_file = fdopen(fd, "a");
 
-    owl_free(open_file);
-    open_file = owl_strdup(filename);
+    g_free(open_file);
+    open_file = g_strdup(filename);
   }
   return g->debug_file;
 }
+
+const char *owl_global_get_kill_buffer(owl_global *g) {
+  return g->kill_buffer;
+}
+
+void owl_global_set_kill_buffer(owl_global *g, const char *kill, int len) {
+  g_free(g->kill_buffer);
+  g->kill_buffer = g_strndup(kill, len);
+}
+
+void owl_global_add_interrupt(owl_global *g) {
+  /* TODO: This can almost certainly be done with atomic
+   * operations. Whatever. */
+  g_mutex_lock(g->interrupt_lock);
+  g->interrupt_count++;
+  g_mutex_unlock(g->interrupt_lock);
+}
+
+bool owl_global_take_interrupt(owl_global *g) {
+  bool ans = false;
+  g_mutex_lock(g->interrupt_lock);
+  if (g->interrupt_count > 0) {
+    ans = true;
+    g->interrupt_count--;
+  }
+  g_mutex_unlock(g->interrupt_lock);
+  return ans;
+}
