@@ -4,10 +4,12 @@
 #undef WINDOW
 
 #include <unistd.h>
+#include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 
 #undef instr
-#include <curses.h>
+#include <ncursesw/curses.h>
 
 owl_global g;
 
@@ -22,6 +24,7 @@ int owl_obarray_regtest(void);
 int owl_editwin_regtest(void);
 int owl_fmtext_regtest(void);
 int owl_smartfilter_regtest(void);
+int owl_history_regtest(void);
 
 extern void owl_perl_xs_init(pTHX);
 
@@ -31,6 +34,7 @@ int main(int argc, char **argv, char **env)
   FILE *wnull;
   char *perlerr;
   int status = 0;
+  SCREEN *screen;
 
   if (argc <= 1) {
     fprintf(stderr, "Usage: %s --builtin|TEST.t|-le CODE\n", argv[0]);
@@ -40,7 +44,7 @@ int main(int argc, char **argv, char **env)
   /* initialize a fake ncurses, detached from std{in,out} */
   wnull = fopen("/dev/null", "w");
   rnull = fopen("/dev/null", "r");
-  newterm("xterm", wnull, rnull);
+  screen = newterm("xterm", wnull, rnull);
   /* initialize global structures */
   owl_global_init(&g);
 
@@ -90,6 +94,7 @@ int main(int argc, char **argv, char **env)
   perl_free(owl_global_get_perlinterp(&g));
   /* probably not necessary, but tear down the screen */
   endwin();
+  delscreen(screen);
   fclose(rnull);
   fclose(wnull);
   return status;
@@ -109,6 +114,7 @@ int owl_regtest(void) {
   numfailures += owl_editwin_regtest();
   numfailures += owl_fmtext_regtest();
   numfailures += owl_smartfilter_regtest();
+  numfailures += owl_history_regtest();
   if (numfailures) {
       fprintf(stderr, "# *** WARNING: %d failures total\n", numfailures);
   }
@@ -119,13 +125,14 @@ int owl_regtest(void) {
 
 #define FAIL_UNLESS(desc,pred) do { int __pred = (pred);                \
     numtests++;                                                         \
-    printf("%s %s", (__pred)?"ok":(numfailed++,"not ok"), desc);        \
+    printf("%s %d %s", (__pred) ? "ok" : (numfailed++, "not ok"), numtests, desc); \
     if(!(__pred)) printf("\t(%s:%d)", __FILE__, __LINE__); printf("%c", '\n'); } while(0)
 
 
 int owl_util_regtest(void)
 {
   int numfailed=0;
+  char *s, *path, *home;
 
   printf("# BEGIN testing owl_util\n");
 
@@ -133,11 +140,9 @@ int owl_util_regtest(void)
     do {                                                 \
       char *__value = (expr);                            \
       FAIL_UNLESS((desc), !strcmp((expected), __value)); \
-      owl_free(__value);                                 \
+      g_free(__value);                                 \
     } while (0)
 
-  CHECK_STR_AND_FREE("owl_util_substitute 1", "foo",
-	             owl_text_substitute("foo", "", "Y"));
   CHECK_STR_AND_FREE("owl_text_substitute 2", "fYZYZ",
                      owl_text_substitute("foo", "o", "YZ"));
   CHECK_STR_AND_FREE("owl_text_substitute 3", "foo",
@@ -185,10 +190,12 @@ int owl_util_regtest(void)
       FAIL_UNLESS(desc, !strcmp(quoted, __quoted));	\
       __argv = owl_parseline(__quoted, &__argc);	\
       FAIL_UNLESS(desc " - arg count", __argc == 1);	\
+      FAIL_UNLESS(desc " - null-terminated",		\
+		  __argv[__argc] == NULL);		\
       FAIL_UNLESS(desc " - parsed",			\
 		  !strcmp(__argv[0], unquoted));	\
-      owl_parse_delete(__argv, __argc);			\
-      owl_free(__quoted);				\
+      g_strfreev(__argv);               		\
+      g_free(__quoted);				\
     } while (0)
 
   CHECK_QUOTING("boring text", "mango", "mango");
@@ -214,11 +221,82 @@ int owl_util_regtest(void)
 		"\"'\"'\""
 		"\"");
 
-  GString *g = g_string_new("");
-  owl_string_appendf_quoted(g, "%q foo %q%q %s %", "hello", "world is", "can't");
+  GString *quoted = g_string_new("");
+  owl_string_appendf_quoted(quoted, "%q foo %q%q %s %", "hello", "world is", "can't");
   FAIL_UNLESS("owl_string_appendf",
-              !strcmp(g_string_free(g, false),
-                      "hello foo 'world is'\"can't\" %s %"));
+              !strcmp(quoted->str, "hello foo 'world is'\"can't\" %s %"));
+  g_string_free(quoted, true);
+
+
+  s = owl_util_baseclass("barnowl");
+  FAIL_UNLESS("baseclass barnowl", !strcmp("barnowl", s));
+  g_free(s);
+  s = owl_util_baseclass("unbarnowl");
+  FAIL_UNLESS("baseclass unbarnowl", !strcmp("barnowl", s));
+  g_free(s);
+  s = owl_util_baseclass("unununbarnowl.d.d");
+  FAIL_UNLESS("baseclass unununbarnowl.d.d", !strcmp("barnowl", s));
+  g_free(s);
+  s = owl_util_baseclass("ununun.d.d");
+  FAIL_UNLESS("baseclass ununun.d.d", !strcmp("", s));
+  g_free(s);
+  s = owl_util_baseclass("d.d.d.d");
+  FAIL_UNLESS("baseclass d.d.d.d", !strcmp("d", s));
+  g_free(s);
+  s = owl_util_baseclass("n.d.d.d");
+  FAIL_UNLESS("baseclass n.d.d.d", !strcmp("n", s));
+  g_free(s);
+  s = owl_util_baseclass("ununun.");
+  FAIL_UNLESS("baseclass ununun.", !strcmp(".", s));
+  g_free(s);
+  s = owl_util_baseclass("unununu");
+  FAIL_UNLESS("baseclass unununu", !strcmp("u", s));
+  g_free(s);
+
+
+  s = owl_util_makepath("foo/bar");
+  FAIL_UNLESS("makepath foo/bar", !strcmp("foo/bar", s));
+  g_free(s);
+  s = owl_util_makepath("//foo///bar");
+  FAIL_UNLESS("makepath //foo///bar", !strcmp("/foo/bar", s));
+  g_free(s);
+  s = owl_util_makepath("foo/~//bar/");
+  FAIL_UNLESS("makepath foo/~//bar/", !strcmp("foo/~/bar/", s));
+  g_free(s);
+  s = owl_util_makepath("~thisuserhadreallybetternotexist/foobar/");
+  FAIL_UNLESS("makepath ~thisuserhadreallybetternotexist/foobar/",
+              !strcmp("~thisuserhadreallybetternotexist/foobar/", s));
+  g_free(s);
+
+  home = g_strdup(owl_global_get_homedir(&g));
+  s = owl_util_makepath("~");
+  FAIL_UNLESS("makepath ~", !strcmp(home, s));
+  g_free(s);
+
+  path = g_build_filename(home, "foo/bar/baz", NULL);
+  s = owl_util_makepath("~///foo/bar//baz");
+  FAIL_UNLESS("makepath ~///foo/bar//baz", !strcmp(path, s));
+  g_free(s);
+  g_free(path);
+  g_free(home);
+
+  home = owl_util_homedir_for_user("root");
+  if (home == NULL) {
+    /* Just make some noise so we notice. */
+    home = g_strdup("<WHAT>");
+    fprintf(stderr, "owl_util_homedir_for_user failed");
+  }
+
+  s = owl_util_makepath("~root");
+  FAIL_UNLESS("makepath ~root", !strcmp(home, s));
+  g_free(s);
+
+  path = g_build_filename(home, "foo/bar/baz", NULL);
+  s = owl_util_makepath("~root///foo/bar//baz");
+  FAIL_UNLESS("makepath ~root///foo/bar//baz", !strcmp(path, s));
+  g_free(s);
+  g_free(path);
+  g_free(home);
 
   /* if (numfailed) printf("*** WARNING: failures encountered with owl_util\n"); */
   printf("# END testing owl_util (%d failures)\n", numfailed);
@@ -227,12 +305,13 @@ int owl_util_regtest(void)
 
 int owl_dict_regtest(void) {
   owl_dict d;
-  owl_list l;
+  GPtrArray *l;
   int numfailed=0;
-  char *av="aval", *bv="bval", *cv="cval", *dv="dval";
+  char *av = g_strdup("aval"), *bv = g_strdup("bval"), *cv = g_strdup("cval"),
+    *dv = g_strdup("dval");
 
   printf("# BEGIN testing owl_dict\n");
-  FAIL_UNLESS("create", 0==owl_dict_create(&d));
+  owl_dict_create(&d);
   FAIL_UNLESS("insert b", 0==owl_dict_insert_element(&d, "b", bv, owl_dict_noop_delete));
   FAIL_UNLESS("insert d", 0==owl_dict_insert_element(&d, "d", dv, owl_dict_noop_delete));
   FAIL_UNLESS("insert a", 0==owl_dict_insert_element(&d, "a", av, owl_dict_noop_delete));
@@ -247,17 +326,22 @@ int owl_dict_regtest(void) {
   FAIL_UNLESS("find d (post-removal)", NULL==owl_dict_find_element(&d, "d"));
 
   FAIL_UNLESS("get_size", 3==owl_dict_get_size(&d));
-  FAIL_UNLESS("get_keys", 0==owl_dict_get_keys(&d, &l));
-  FAIL_UNLESS("get_keys result size", 3==owl_list_get_size(&l));
+  l = owl_dict_get_keys(&d);
+  FAIL_UNLESS("get_keys result size", 3 == l->len);
   
   /* these assume the returned keys are sorted */
-  FAIL_UNLESS("get_keys result val",0==strcmp("a",owl_list_get_element(&l,0)));
-  FAIL_UNLESS("get_keys result val",0==strcmp("b",owl_list_get_element(&l,1)));
-  FAIL_UNLESS("get_keys result val",0==strcmp("c",owl_list_get_element(&l,2)));
+  FAIL_UNLESS("get_keys result val", 0 == strcmp("a", l->pdata[0]));
+  FAIL_UNLESS("get_keys result val", 0 == strcmp("b", l->pdata[1]));
+  FAIL_UNLESS("get_keys result val", 0 == strcmp("c", l->pdata[2]));
 
-  owl_list_cleanup(&l, owl_free);
+  owl_ptr_array_free(l, g_free);
   owl_dict_cleanup(&d, NULL);
 
+  g_free(av);
+  g_free(bv);
+  g_free(cv);
+  g_free(dv);
+
   /*  if (numfailed) printf("*** WARNING: failures encountered with owl_dict\n"); */
   printf("# END testing owl_dict (%d failures)\n", numfailed);
   return(numfailed);
@@ -265,60 +349,77 @@ int owl_dict_regtest(void) {
 
 int owl_variable_regtest(void) {
   owl_vardict vd;
+  owl_variable *var;
   int numfailed=0;
-  char buf[1024];
+  char *value;
   const void *v;
 
   printf("# BEGIN testing owl_variable\n");
   FAIL_UNLESS("setup", 0==owl_variable_dict_setup(&vd));
 
-  FAIL_UNLESS("get bool", 0==owl_variable_get_bool(&vd,"rxping"));
-  FAIL_UNLESS("get bool (no such)", -1==owl_variable_get_bool(&vd,"mumble"));
-  FAIL_UNLESS("get bool as string 1", 0==owl_variable_get_tostring(&vd,"rxping", buf, 1024));
-  FAIL_UNLESS("get bool as string 2", 0==strcmp(buf,"off"));
-  FAIL_UNLESS("set bool 1", 0==owl_variable_set_bool_on(&vd,"rxping"));
-  FAIL_UNLESS("get bool 2", 1==owl_variable_get_bool(&vd,"rxping"));
-  FAIL_UNLESS("set bool 3", 0==owl_variable_set_fromstring(&vd,"rxping","off",0,0));
-  FAIL_UNLESS("get bool 4", 0==owl_variable_get_bool(&vd,"rxping"));
-  FAIL_UNLESS("set bool 5", -1==owl_variable_set_fromstring(&vd,"rxping","xxx",0,0));
-  FAIL_UNLESS("get bool 6", 0==owl_variable_get_bool(&vd,"rxping"));
+  FAIL_UNLESS("get bool var", NULL != (var = owl_variable_get_var(&vd, "rxping")));
+  FAIL_UNLESS("get bool", 0 == owl_variable_get_bool(var));
+  FAIL_UNLESS("get bool (no such)", NULL == owl_variable_get_var(&vd, "mumble"));
+  FAIL_UNLESS("get bool as string",
+	      !strcmp((value = owl_variable_get_tostring(var)), "off"));
+  g_free(value);
+  FAIL_UNLESS("set bool 1", 0 == owl_variable_set_bool_on(var));
+  FAIL_UNLESS("get bool 2", 1 == owl_variable_get_bool(var));
+  FAIL_UNLESS("set bool 3", 0 == owl_variable_set_fromstring(var, "off", 0));
+  FAIL_UNLESS("get bool 4", 0 == owl_variable_get_bool(var));
+  FAIL_UNLESS("set bool 5", -1 == owl_variable_set_fromstring(var, "xxx", 0));
+  FAIL_UNLESS("get bool 6", 0 == owl_variable_get_bool(var));
 
 
-  FAIL_UNLESS("get string", 0==strcmp("~/zlog/people", owl_variable_get_string(&vd,"logpath")));
-  FAIL_UNLESS("set string 7", 0==owl_variable_set_string(&vd,"logpath","whee"));
-  FAIL_UNLESS("get string", 0==strcmp("whee", owl_variable_get_string(&vd,"logpath")));
+  FAIL_UNLESS("get string var", NULL != (var = owl_variable_get_var(&vd, "logpath")));
+  FAIL_UNLESS("get string", 0 == strcmp("~/zlog/people", owl_variable_get_string(var)));
+  FAIL_UNLESS("set string 7", 0 == owl_variable_set_string(var, "whee"));
+  FAIL_UNLESS("get string", !strcmp("whee", owl_variable_get_string(var)));
 
-  FAIL_UNLESS("get int", 8==owl_variable_get_int(&vd,"typewinsize"));
-  FAIL_UNLESS("get int (no such)", -1==owl_variable_get_int(&vd,"mmble"));
-  FAIL_UNLESS("get int as string 1", 0==owl_variable_get_tostring(&vd,"typewinsize", buf, 1024));
-  FAIL_UNLESS("get int as string 2", 0==strcmp(buf,"8"));
-  FAIL_UNLESS("set int 1", 0==owl_variable_set_int(&vd,"typewinsize",12));
-  FAIL_UNLESS("get int 2", 12==owl_variable_get_int(&vd,"typewinsize"));
-  FAIL_UNLESS("set int 1b", -1==owl_variable_set_int(&vd,"typewinsize",-3));
-  FAIL_UNLESS("get int 2b", 12==owl_variable_get_int(&vd,"typewinsize"));
-  FAIL_UNLESS("set int 3", 0==owl_variable_set_fromstring(&vd,"typewinsize","9",0,0));
-  FAIL_UNLESS("get int 4", 9==owl_variable_get_int(&vd,"typewinsize"));
-  FAIL_UNLESS("set int 5", -1==owl_variable_set_fromstring(&vd,"typewinsize","xxx",0,0));
-  FAIL_UNLESS("set int 6", -1==owl_variable_set_fromstring(&vd,"typewinsize","",0,0));
-  FAIL_UNLESS("get int 7", 9==owl_variable_get_int(&vd,"typewinsize"));
+  FAIL_UNLESS("get int var", NULL != (var = owl_variable_get_var(&vd, "typewinsize")));
+  FAIL_UNLESS("get int", 8 == owl_variable_get_int(var));
+  FAIL_UNLESS("get int (no such)", NULL == owl_variable_get_var(&vd, "mumble"));
+  FAIL_UNLESS("get int as string",
+	      !strcmp((value = owl_variable_get_tostring(var)), "8"));
+  g_free(value);
+  FAIL_UNLESS("set int 1", 0 == owl_variable_set_int(var, 12));
+  FAIL_UNLESS("get int 2", 12 == owl_variable_get_int(var));
+  FAIL_UNLESS("set int 1b", -1 == owl_variable_set_int(var, -3));
+  FAIL_UNLESS("get int 2b", 12 == owl_variable_get_int(var));
+  FAIL_UNLESS("set int 3", 0 == owl_variable_set_fromstring(var, "9", 0));
+  FAIL_UNLESS("get int 4", 9 == owl_variable_get_int(var));
+  FAIL_UNLESS("set int 5", -1 == owl_variable_set_fromstring(var, "xxx", 0));
+  FAIL_UNLESS("set int 6", -1 == owl_variable_set_fromstring(var, "", 0));
+  FAIL_UNLESS("get int 7", 9 == owl_variable_get_int(var));
 
   owl_variable_dict_newvar_string(&vd, "stringvar", "", "", "testval");
-  FAIL_UNLESS("get new string var", NULL != (v = owl_variable_get(&vd, "stringvar", OWL_VARIABLE_STRING)));
-  FAIL_UNLESS("get new string val", !strcmp("testval", owl_variable_get_string(&vd, "stringvar")));
-  owl_variable_set_string(&vd, "stringvar", "new val");
-  FAIL_UNLESS("update string val", !strcmp("new val", owl_variable_get_string(&vd, "stringvar")));
+  FAIL_UNLESS("get new string var", NULL != (var = owl_variable_get_var(&vd, "stringvar")));
+  FAIL_UNLESS("get new string var", NULL != (v = owl_variable_get(var)));
+  FAIL_UNLESS("get new string val", !strcmp("testval", owl_variable_get_string(var)));
+  owl_variable_set_string(var, "new val");
+  FAIL_UNLESS("update string val", !strcmp("new val", owl_variable_get_string(var)));
 
   owl_variable_dict_newvar_int(&vd, "intvar", "", "", 47);
-  FAIL_UNLESS("get new int var", NULL != (v = owl_variable_get(&vd, "intvar", OWL_VARIABLE_INT)));
-  FAIL_UNLESS("get new int val", 47 == owl_variable_get_int(&vd, "intvar"));
-  owl_variable_set_int(&vd, "intvar", 17);
-  FAIL_UNLESS("update bool val", 17 == owl_variable_get_int(&vd, "intvar"));
+  FAIL_UNLESS("get new int var", NULL != (var = owl_variable_get_var(&vd, "intvar")));
+  FAIL_UNLESS("get new int var", NULL != (v = owl_variable_get(var)));
+  FAIL_UNLESS("get new int val", 47 == owl_variable_get_int(var));
+  owl_variable_set_int(var, 17);
+  FAIL_UNLESS("update int val", 17 == owl_variable_get_int(var));
 
   owl_variable_dict_newvar_bool(&vd, "boolvar", "", "", 1);
-  FAIL_UNLESS("get new bool var", NULL != (v = owl_variable_get(&vd, "boolvar", OWL_VARIABLE_BOOL)));
-  FAIL_UNLESS("get new bool val", owl_variable_get_bool(&vd, "boolvar"));
-  owl_variable_set_bool_off(&vd, "boolvar");
-  FAIL_UNLESS("update string val", !owl_variable_get_bool(&vd, "boolvar"));
+  FAIL_UNLESS("get new bool var", NULL != (var = owl_variable_get_var(&vd, "boolvar")));
+  FAIL_UNLESS("get new bool var", NULL != (v = owl_variable_get(var)));
+  FAIL_UNLESS("get new bool val", owl_variable_get_bool(var));
+  owl_variable_set_bool_off(var);
+  FAIL_UNLESS("update bool val", !owl_variable_get_bool(var));
+
+  owl_variable_dict_newvar_string(&vd, "nullstringvar", "", "", NULL);
+  FAIL_UNLESS("get new string (NULL) var", NULL != (var = owl_variable_get_var(&vd, "nullstringvar")));
+  FAIL_UNLESS("get string (NULL)", NULL == (value = owl_variable_get_tostring(var)));
+  g_free(value);
+  var = owl_variable_get_var(&vd, "zsigproc");
+  FAIL_UNLESS("get string (NULL) 2", NULL == (value = owl_variable_get_tostring(var)));
+  g_free(value);
 
   owl_variable_dict_cleanup(&vd);
 
@@ -554,14 +655,14 @@ int owl_fmtext_regtest(void) {
   str = owl_fmtext_print_plain(&fm2);
   FAIL_UNLESS("lines corrected truncated",
 	      str && !strcmp(str, "1\n2\n3\n"));
-  owl_free(str);
+  g_free(str);
 
   owl_fmtext_clear(&fm2);
   owl_fmtext_truncate_lines(&fm1, 1, 5, &fm2);
   str = owl_fmtext_print_plain(&fm2);
   FAIL_UNLESS("lines corrected truncated",
 	      str && !strcmp(str, "1\n2\n3\n4\n"));
-  owl_free(str);
+  g_free(str);
 
   /* Test owl_fmtext_truncate_cols. */
   owl_fmtext_clear(&fm1);
@@ -576,7 +677,7 @@ int owl_fmtext_regtest(void) {
               str && !strcmp(str, "567890"
                                   "56789\n"
                                   "567890"));
-  owl_free(str);
+  g_free(str);
 
   owl_fmtext_clear(&fm1);
   owl_fmtext_append_normal(&fm1, "12\t1234");
@@ -589,7 +690,7 @@ int owl_fmtext_regtest(void) {
   FAIL_UNLESS("columns correctly truncated",
               str && !strcmp(str, "    123456"
                                   "5678      "));
-  owl_free(str);
+  g_free(str);
 
   /* Test owl_fmtext_expand_tabs. */
   owl_fmtext_clear(&fm1);
@@ -602,7 +703,7 @@ int owl_fmtext_regtest(void) {
   FAIL_UNLESS("tabs corrected expanded",
               str && !strcmp(str, "12      1234567 1\n"
                                   "12345678        1"));
-  owl_free(str);
+  g_free(str);
 
   owl_fmtext_clear(&fm2);
   owl_fmtext_expand_tabs(&fm1, &fm2, 1);
@@ -611,7 +712,7 @@ int owl_fmtext_regtest(void) {
   FAIL_UNLESS("tabs corrected expanded",
               str && !strcmp(str, "12     1234567 1\n"
                                   "12345678       1"));
-  owl_free(str);
+  g_free(str);
 
   /* Test owl_fmtext_search. */
   owl_fmtext_clear(&fm1);
@@ -680,7 +781,7 @@ static int owl_smartfilter_test_equals(const char *filtname, const char *expecte
   }
 
  out:
-  owl_free(filtstr);
+  g_free(filtstr);
   return failed;
 }
 
@@ -705,7 +806,7 @@ static int owl_classinstfilt_test(const char *c, const char *i, int related, con
   }
   if (filtname)
     owl_global_remove_filter(&g, filtname);
-  owl_free(filtname);
+  g_free(filtname);
   return failed;
 }
 
@@ -729,7 +830,7 @@ static int owl_zuserfilt_test(const char *longuser, const char *expected) {
   }
   if (filtname)
     owl_global_remove_filter(&g, filtname);
-  owl_free(filtname);
+  g_free(filtname);
   return failed;
 }
 
@@ -792,3 +893,91 @@ int owl_smartfilter_regtest(void) {
 
   return numfailed;
 }
+
+int owl_history_regtest(void)
+{
+  int numfailed = 0;
+  int i;
+  owl_history h;
+
+  printf("# BEGIN testing owl_history\n");
+  owl_history_init(&h);
+
+  /* Operations on empty history. */
+  FAIL_UNLESS("prev NULL", owl_history_get_prev(&h) == NULL);
+  FAIL_UNLESS("next NULL", owl_history_get_next(&h) == NULL);
+  FAIL_UNLESS("untouched", !owl_history_is_touched(&h));
+
+  /* Insert a few records. */
+  owl_history_store(&h, "a", false);
+  owl_history_store(&h, "b", false);
+  owl_history_store(&h, "c", false);
+  owl_history_store(&h, "d", true);
+
+  /* Walk up and down the history a bit. */
+  FAIL_UNLESS("untouched", !owl_history_is_touched(&h));
+  FAIL_UNLESS("prev c", strcmp(owl_history_get_prev(&h), "c") == 0);
+  FAIL_UNLESS("touched", owl_history_is_touched(&h));
+  FAIL_UNLESS("next d", strcmp(owl_history_get_next(&h), "d") == 0);
+  FAIL_UNLESS("untouched", !owl_history_is_touched(&h));
+  FAIL_UNLESS("next NULL", owl_history_get_next(&h) == NULL);
+  FAIL_UNLESS("prev c", strcmp(owl_history_get_prev(&h), "c") == 0);
+  FAIL_UNLESS("prev b", strcmp(owl_history_get_prev(&h), "b") == 0);
+  FAIL_UNLESS("prev a", strcmp(owl_history_get_prev(&h), "a") == 0);
+  FAIL_UNLESS("prev NULL", owl_history_get_prev(&h) == NULL);
+
+  /* Now insert something. It should reset and blow away 'd'. */
+  owl_history_store(&h, "e", false);
+  FAIL_UNLESS("untouched", !owl_history_is_touched(&h));
+  FAIL_UNLESS("next NULL", owl_history_get_next(&h) == NULL);
+  FAIL_UNLESS("prev c", strcmp(owl_history_get_prev(&h), "c") == 0);
+  FAIL_UNLESS("touched", owl_history_is_touched(&h));
+  FAIL_UNLESS("next e", strcmp(owl_history_get_next(&h), "e") == 0);
+  FAIL_UNLESS("untouched", !owl_history_is_touched(&h));
+
+  /* Lines get de-duplicated on insert. */
+  owl_history_store(&h, "e", false);
+  owl_history_store(&h, "e", false);
+  owl_history_store(&h, "e", false);
+  FAIL_UNLESS("prev c", strcmp(owl_history_get_prev(&h), "c") == 0);
+  FAIL_UNLESS("next e", strcmp(owl_history_get_next(&h), "e") == 0);
+
+  /* But a partial is not deduplicated, as it'll go away soon. */
+  owl_history_store(&h, "e", true);
+  FAIL_UNLESS("prev e", strcmp(owl_history_get_prev(&h), "e") == 0);
+  FAIL_UNLESS("prev c", strcmp(owl_history_get_prev(&h), "c") == 0);
+  FAIL_UNLESS("next e", strcmp(owl_history_get_next(&h), "e") == 0);
+  FAIL_UNLESS("next e", strcmp(owl_history_get_next(&h), "e") == 0);
+
+  /* Reset moves to the front... */
+  owl_history_store(&h, "f", true);
+  FAIL_UNLESS("prev e", strcmp(owl_history_get_prev(&h), "e") == 0);
+  FAIL_UNLESS("prev c", strcmp(owl_history_get_prev(&h), "c") == 0);
+  owl_history_reset(&h);
+  FAIL_UNLESS("untouched", !owl_history_is_touched(&h));
+  /* ...and destroys any pending partial entry... */
+  FAIL_UNLESS("prev c", strcmp(owl_history_get_prev(&h), "c") == 0);
+  FAIL_UNLESS("prev b", strcmp(owl_history_get_prev(&h), "b") == 0);
+  /* ...but not non-partial ones. */
+  owl_history_reset(&h);
+  FAIL_UNLESS("untouched", !owl_history_is_touched(&h));
+
+  /* Finally, check we are bounded by OWL_HISTORYSIZE. */
+  for (i = 0; i < OWL_HISTORYSIZE; i++) {
+    char *string = g_strdup_printf("mango%d", i);
+    owl_history_store(&h, string, false);
+    g_free(string);
+  }
+  /* The OWL_HISTORYSIZE'th prev gets NULL. */
+  for (i = OWL_HISTORYSIZE - 2; i >= 0; i--) {
+    char *string = g_strdup_printf("mango%d", i);
+    FAIL_UNLESS("prev mango_N", strcmp(owl_history_get_prev(&h), string) == 0);
+    g_free(string);
+  }
+  FAIL_UNLESS("prev NULL", owl_history_get_prev(&h) == NULL);
+
+  owl_history_cleanup(&h);
+
+  printf("# END testing owl_history (%d failures)\n", numfailed);
+  return numfailed;
+}
