@@ -12,36 +12,6 @@
 #include <glib/gstdio.h>
 #include <glib-object.h>
 
-char **atokenize(const char *buffer, const char *sep, int *i)
-{
-  /* each element of return must be freed by user */
-  char **args;
-  char *workbuff, *foo;
-  int done=0, first=1, count=0;
-
-  workbuff = owl_strdup(buffer);
-
-  args=NULL;
-  while (!done) {
-    if (first) {
-      first=0;
-      foo=strtok(workbuff, sep);
-    } else {
-      foo=strtok(NULL, sep);
-    }
-    if (foo==NULL) {
-      done=1;
-    } else {
-      args=owl_realloc(args, sizeof(char *) * (count+1));
-      args[count] = owl_strdup(foo);
-      count++;
-    }
-  }
-  *i=count;
-  owl_free(workbuff);
-  return(args);
-}
-
 const char *skiptokens(const char *buff, int n) {
   /* skips n tokens and returns where that would be. */
   char quote = 0;
@@ -61,112 +31,101 @@ const char *skiptokens(const char *buff, int n) {
   return buff;
 }
 
+CALLER_OWN char *owl_util_homedir_for_user(const char *name)
+{
+  int err;
+  struct passwd pw_buf;
+  struct passwd *pw;
+
+  char *pw_strbuf, *ret;
+  long pw_strbuf_len = sysconf(_SC_GETPW_R_SIZE_MAX);
+  if (pw_strbuf_len < 0) {
+    /* If we really hate ourselves, we can be fancy and loop until we stop
+     * getting ERANGE. For now just pick a random number. */
+    owl_function_error("owl_util_homedir_for_user: Could not get _SC_GETPW_R_SIZE_MAX");
+    pw_strbuf_len = 16384;
+  }
+  pw_strbuf = g_new0(char, pw_strbuf_len);
+  err = getpwnam_r(name, &pw_buf, pw_strbuf, pw_strbuf_len, &pw);
+  if (err) {
+    owl_function_error("getpwuid_r: %s", strerror(err));
+    /* Fall through; pw will be NULL. */
+  }
+  ret = pw ? g_strdup(pw->pw_dir) : NULL;
+  g_free(pw_strbuf);
+  return ret;
+}
+
 /* Return a "nice" version of the path.  Tilde expansion is done, and
  * duplicate slashes are removed.  Caller must free the return.
  */
-char *owl_util_makepath(const char *in)
+CALLER_OWN char *owl_util_makepath(const char *in)
 {
-  int i, j, x;
-  char *out, user[MAXPATHLEN];
-  struct passwd *pw;
+  char *out;
+  int i, j;
+  if (in[0] == '~') {
+    /* Attempt tilde-expansion of the first component. Get the
+       tilde-prefix, which goes up to the next slash. */
+    const char *end = strchr(in + 1, '/');
+    if (end == NULL)
+      end = in + strlen(in);
 
-  out=owl_malloc(MAXPATHLEN+1);
-  out[0]='\0';
-  j=strlen(in);
-  x=0;
-  for (i=0; i<j; i++) {
-    if (in[i]=='~') {
-      if ( (i==(j-1)) ||          /* last character */
-	   (in[i+1]=='/') ) {     /* ~/ */
-	/* use my homedir */
-	pw=getpwuid(getuid());
-	if (!pw) {
-	  out[x]=in[i];
-	} else {
-	  out[x]='\0';
-	  strcat(out, pw->pw_dir);
-	  x+=strlen(pw->pw_dir);
-	}
-      } else {
-	/* another user homedir */
-	int a, b;
-	b=0;
-	for (a=i+1; i<j; a++) {
-	  if (in[a]==' ' || in[a]=='/') {
-	    break;
-	  } else {
-	    user[b]=in[a];
-	    i++;
-	    b++;
-	  }
-	}
-	user[b]='\0';
-	pw=getpwnam(user);
-	if (!pw) {
-	  out[x]=in[i];
-	} else {
-	  out[x]='\0';
-	  strcat(out, pw->pw_dir);
-	  x+=strlen(pw->pw_dir);
-	}
-      }
-    } else if (in[i]=='/') {
-      /* check for a double / */
-      if (i<(j-1) && (in[i+1]=='/')) {
-	/* do nothing */
-      } else {
-	out[x]=in[i];
-	x++;
-      }
+    /* Patch together a new path. Replace the ~ and tilde-prefix with
+       the homedir, if available. */
+    if (end == in + 1) {
+      /* My home directory. Use the one in owl_global for consistency with
+       * owl_zephyr_dotfile. */
+      out = g_strconcat(owl_global_get_homedir(&g), end, NULL);
     } else {
-      out[x]=in[i];
-      x++;
+      /* Someone else's home directory. */
+      char *user = g_strndup(in + 1, end - (in + 1));
+      char *home = owl_util_homedir_for_user(user);
+      if (home) {
+        out = g_strconcat(home, end, NULL);
+      } else {
+        out = g_strdup(in);
+      }
+      g_free(home);
+      g_free(user);
     }
+  } else {
+    out = g_strdup(in);
   }
-  out[x]='\0';
-  return(out);
+
+  /* And a quick pass to remove duplicate slashes. */
+  for (i = j = 0; out[i] != '\0'; i++) {
+    if (out[i] != '/' || i == 0 || out[i-1] != '/')
+      out[j++] = out[i];
+  }
+  out[j] = '\0';
+  return out;
 }
 
-void atokenize_delete(char **tok, int nels)
+void owl_ptr_array_free(GPtrArray *array, GDestroyNotify element_free_func)
 {
-  int i;
-  for (i=0; i<nels; i++) {
-    owl_free(tok[i]);
-  }
-  owl_free(tok);
+  /* TODO: when we move to requiring glib 2.22+, use
+   * g_ptr_array_new_with_free_func instead. */
+  if (element_free_func)
+    g_ptr_array_foreach(array, (GFunc)element_free_func, NULL);
+  g_ptr_array_free(array, true);
 }
 
-
-void owl_parse_delete(char **argv, int argc)
+/* Break a command line up into argv, argc.  The caller must free
+   the returned values with g_strfreev.  If there is an error argc will be set
+   to -1, argv will be NULL and the caller does not need to free anything. The
+   returned vector is NULL-terminated. */
+CALLER_OWN char **owl_parseline(const char *line, int *argc)
 {
-  int i;
-
-  if (!argv) return;
-  
-  for (i=0; i<argc; i++) {
-    if (argv[i]) owl_free(argv[i]);
-  }
-  owl_free(argv);
-}
-
-char **owl_parseline(const char *line, int *argc)
-{
-  /* break a command line up into argv, argc.  The caller must free
-     the returned values.  If there is an error argc will be set to
-     -1, argv will be NULL and the caller does not need to free
-     anything */
-
-  char **argv;
+  GPtrArray *argv;
   int i, len, between=1;
-  char *curarg;
+  GString *curarg;
   char quote;
 
-  argv=owl_malloc(sizeof(char *));
+  argv = g_ptr_array_new();
   len=strlen(line);
-  curarg=owl_malloc(len+10);
-  strcpy(curarg, "");
+  curarg = g_string_new("");
   quote='\0';
-  *argc=0;
+  if (argc) *argc=0;
   for (i=0; i<len+1; i++) {
     /* find the first real character */
     if (between) {
@@ -194,44 +153,41 @@ char **owl_parseline(const char *line, int *argc)
       }
 
       /* if another type of quote is open then treat this as a literal */
-      curarg[strlen(curarg)+1]='\0';
-      curarg[strlen(curarg)]=line[i];
+      g_string_append_c(curarg, line[i]);
       continue;
     }
 
     /* if it's not a space or end of command, then use it */
     if (line[i]!=' ' && line[i]!='\t' && line[i]!='\n' && line[i]!='\0') {
-      curarg[strlen(curarg)+1]='\0';
-      curarg[strlen(curarg)]=line[i];
+      g_string_append_c(curarg, line[i]);
       continue;
     }
 
     /* otherwise, if we're not in quotes, add the whole argument */
     if (quote=='\0') {
       /* add the argument */
-      argv=owl_realloc(argv, sizeof(char *)*((*argc)+1));
-      argv[*argc] = owl_strdup(curarg);
-      *argc=*argc+1;
-      strcpy(curarg, "");
+      g_ptr_array_add(argv, g_string_free(curarg, false));
+      curarg = g_string_new("");
       between=1;
       continue;
     }
 
     /* if it is a space and we're in quotes, then use it */
-    curarg[strlen(curarg)+1]='\0';
-    curarg[strlen(curarg)]=line[i];
+    g_string_append_c(curarg, line[i]);
   }
 
-  owl_free(curarg);
+  if (argc) *argc = argv->len;
+  g_ptr_array_add(argv, NULL);
+  g_string_free(curarg, true);
 
   /* check for unbalanced quotes */
   if (quote!='\0') {
-    owl_parse_delete(argv, *argc);
-    *argc=-1;
+    owl_ptr_array_free(argv, g_free);
+    if (argc) *argc = -1;
     return(NULL);
   }
 
-  return(argv);
+  return (char**)g_ptr_array_free(argv, false);
 }
 
 /* Appends a quoted version of arg suitable for placing in a
@@ -268,7 +224,7 @@ void owl_string_append_quoted_arg(GString *buf, const char *arg)
 
 /*
  * Appends 'tmpl' to 'buf', replacing any instances of '%q' with arguments from
- * the varargs provided, quoting them to be safe for placing in a barnowl
+ * the varargs provided, quoting them to be safe for placing in a BarnOwl
  * command line.
  */
 void owl_string_appendf_quoted(GString *buf, const char *tmpl, ...)
@@ -298,7 +254,7 @@ void owl_string_vappendf_quoted(GString *buf, const char *tmpl, va_list ap)
   g_string_append(buf, last);
 }
 
-char *owl_string_build_quoted(const char *tmpl, ...)
+CALLER_OWN char *owl_string_build_quoted(const char *tmpl, ...)
 {
   GString *buf = g_string_new("");
   va_list ap;
@@ -309,8 +265,8 @@ char *owl_string_build_quoted(const char *tmpl, ...)
 }
 
 /* Returns a quoted version of arg suitable for placing in a
- * command-line. Result should be freed with owl_free. */
-char *owl_arg_quote(const char *arg)
+ * command-line. Result should be freed with g_free. */
+CALLER_OWN char *owl_arg_quote(const char *arg)
 {
   GString *buf = g_string_new("");;
   owl_string_append_quoted_arg(buf, arg);
@@ -318,7 +274,7 @@ char *owl_arg_quote(const char *arg)
 }
 
 /* caller must free the return */
-char *owl_util_minutes_to_timestr(int in)
+CALLER_OWN char *owl_util_minutes_to_timestr(int in)
 {
   int days, hours;
   long run;
@@ -332,48 +288,13 @@ char *owl_util_minutes_to_timestr(int in)
   run-=hours*60;
 
   if (days>0) {
-    out=owl_sprintf("%i d %2.2i:%2.2li", days, hours, run);
+    out=g_strdup_printf("%i d %2.2i:%2.2li", days, hours, run);
   } else {
-    out=owl_sprintf("    %2.2i:%2.2li", hours, run);
+    out=g_strdup_printf("    %2.2i:%2.2li", hours, run);
   }
   return(out);
 }
 
-/* hooks for doing memory allocation et. al. in owl */
-
-void *owl_malloc(size_t size)
-{
-  return(g_malloc(size));
-}
-
-void owl_free(void *ptr)
-{
-  g_free(ptr);
-}
-
-char *owl_strdup(const char *s1)
-{
-  return(g_strdup(s1));
-}
-
-void *owl_realloc(void *ptr, size_t size)
-{
-  return(g_realloc(ptr, size));
-}
-
-/* allocates memory and returns the string or null.
- * caller must free the string. 
- */
-char *owl_sprintf(const char *fmt, ...)
-{
-  va_list ap;
-  char *ret = NULL;
-  va_start(ap, fmt);
-  ret = g_strdup_vprintf(fmt, ap);
-  va_end(ap);
-  return ret;
-}
-
 /* These are in order of their value in owl.h */
 static const struct {
   int number;
@@ -419,21 +340,21 @@ const char *owl_util_color_to_string(int color)
 }
 
 /* Get the default tty name.  Caller must free the return */
-char *owl_util_get_default_tty(void)
+CALLER_OWN char *owl_util_get_default_tty(void)
 {
   const char *tmp;
   char *out;
 
   if (getenv("DISPLAY")) {
-    out=owl_strdup(getenv("DISPLAY"));
+    out=g_strdup(getenv("DISPLAY"));
   } else if ((tmp=ttyname(fileno(stdout)))!=NULL) {
-    out=owl_strdup(tmp);
+    out=g_strdup(tmp);
     if (!strncmp(out, "/dev/", 5)) {
-      owl_free(out);
-      out=owl_strdup(tmp+5);
+      g_free(out);
+      out=g_strdup(tmp+5);
     }
   } else {
-    out=owl_strdup("unknown");
+    out=g_strdup("unknown");
   }
   return(out);
 }
@@ -441,12 +362,12 @@ char *owl_util_get_default_tty(void)
 /* strip leading and trailing new lines.  Caller must free the
  * return.
  */
-char *owl_util_stripnewlines(const char *in)
+CALLER_OWN char *owl_util_stripnewlines(const char *in)
 {
   
   char  *tmp, *ptr1, *ptr2, *out;
 
-  ptr1=tmp=owl_strdup(in);
+  ptr1=tmp=g_strdup(in);
   while (ptr1[0]=='\n') {
     ptr1++;
   }
@@ -456,8 +377,8 @@ char *owl_util_stripnewlines(const char *in)
     ptr2--;
   }
 
-  out=owl_strdup(ptr1);
-  owl_free(tmp);
+  out=g_strdup(ptr1);
+  g_free(tmp);
   return(out);
 }
 
@@ -472,7 +393,7 @@ char *owl_util_stripnewlines(const char *in)
  *
  * Error conditions are the same as g_file_read_link.
  */
-gchar *owl_util_recursive_resolve_link(const char *filename)
+CALLER_OWN gchar *owl_util_recursive_resolve_link(const char *filename)
 {
   gchar *last_path = g_strdup(filename);
   GError *err = NULL;
@@ -491,10 +412,7 @@ gchar *owl_util_recursive_resolve_link(const char *filename)
      * is racy. Whatever. */
     if (!g_path_is_absolute(link_path)) {
       char *last_dir = g_path_get_dirname(last_path);
-      char *tmp = g_build_path(G_DIR_SEPARATOR_S,
-			       last_dir,
-			       link_path,
-			       NULL);
+      char *tmp = g_build_filename(last_dir, link_path, NULL);
       g_free(last_dir);
       g_free(link_path);
       link_path = tmp;
@@ -537,13 +455,13 @@ int owl_util_file_deleteline(const char *filename, const char *line, int backup)
   if (actual_filename == NULL)
     return -1; /* resolving the symlink failed, but we already logged this error */
 
-  newfile = owl_sprintf("%s.new", actual_filename);
+  newfile = g_strdup_printf("%s.new", actual_filename);
   if ((new = fopen(newfile, "w")) == NULL) {
     owl_function_error("Cannot open %s (for writing): %s",
 		       actual_filename, strerror(errno));
-    owl_free(newfile);
+    g_free(newfile);
     fclose(old);
-    free(actual_filename);
+    g_free(actual_filename);
     return -1;
   }
 
@@ -552,9 +470,9 @@ int owl_util_file_deleteline(const char *filename, const char *line, int backup)
 		       actual_filename, strerror(errno));
     unlink(newfile);
     fclose(new);
-    owl_free(newfile);
+    g_free(newfile);
     fclose(old);
-    free(actual_filename);
+    g_free(actual_filename);
     return -1;
   }
 
@@ -563,22 +481,22 @@ int owl_util_file_deleteline(const char *filename, const char *line, int backup)
       fprintf(new, "%s\n", buf);
     else
       numremoved++;
-  owl_free(buf);
+  g_free(buf);
 
   fclose(new);
   fclose(old);
 
   if (backup) {
-    backupfile = owl_sprintf("%s.backup", actual_filename);
+    backupfile = g_strdup_printf("%s.backup", actual_filename);
     unlink(backupfile);
     if (link(actual_filename, backupfile) != 0) {
       owl_function_error("Cannot link %s: %s", backupfile, strerror(errno));
-      owl_free(backupfile);
+      g_free(backupfile);
       unlink(newfile);
-      owl_free(newfile);
+      g_free(newfile);
       return -1;
     }
-    owl_free(backupfile);
+    g_free(backupfile);
   }
 
   if (rename(newfile, actual_filename) != 0) {
@@ -588,30 +506,18 @@ int owl_util_file_deleteline(const char *filename, const char *line, int backup)
   }
 
   unlink(newfile);
-  owl_free(newfile);
+  g_free(newfile);
 
   g_free(actual_filename);
 
   return numremoved;
 }
 
-int owl_util_max(int a, int b)
-{
-  if (a>b) return(a);
-  return(b);
-}
-
-int owl_util_min(int a, int b)
-{
-  if (a<b) return(a);
-  return(b);
-}
-
 /* Return the base class or instance from a zephyr class, by removing
    leading `un' or trailing `.d'.
    The caller is responsible for freeing the allocated string.
 */
-char * owl_util_baseclass(const char * class)
+CALLER_OWN char *owl_util_baseclass(const char *class)
 {
   char *start, *end;
 
@@ -619,7 +525,7 @@ char * owl_util_baseclass(const char * class)
     class += 2;
   }
 
-  start = owl_strdup(class);
+  start = g_strdup(class);
   end = start + strlen(start) - 1;
   while(end > start && *end == 'd' && *(end-1) == '.') {
     end -= 2;
@@ -646,13 +552,13 @@ const char * owl_get_bindir(void)
 }
 
 /* Strips format characters from a valid utf-8 string. Returns the
-   empty string if 'in' does not validate. */
-char * owl_strip_format_chars(const char *in)
+   empty string if 'in' does not validate.  Caller must free the return. */
+CALLER_OWN char *owl_strip_format_chars(const char *in)
 {
   char *r;
   if (g_utf8_validate(in, -1, NULL)) {
     const char *s, *p;
-    r = owl_malloc(strlen(in)+1);
+    r = g_new(char, strlen(in)+1);
     r[0] = '\0';
     s = in;
     p = strchr(s, OWL_FMTEXT_UC_STARTBYTE_UTF8);
@@ -675,7 +581,7 @@ char * owl_strip_format_chars(const char *in)
     if (s) strcat(r,s);
   }
   else {
-    r = owl_strdup("");
+    r = g_strdup("");
   }
   return r;
 }
@@ -684,8 +590,9 @@ char * owl_strip_format_chars(const char *in)
  * the caller to specify an alternative in the future. We also strip
  * out characters in Unicode Plane 16, as we use that plane internally
  * for formatting.
+ * Caller must free the return.
  */
-char * owl_validate_or_convert(const char *in)
+CALLER_OWN char *owl_validate_or_convert(const char *in)
 {
   if (g_utf8_validate(in, -1, NULL)) {
     return owl_strip_format_chars(in);
@@ -699,14 +606,15 @@ char * owl_validate_or_convert(const char *in)
 /*
  * Validate 'in' as UTF-8, and either return a copy of it, or an empty
  * string if it is invalid utf-8.
+ * Caller must free the return.
  */
-char * owl_validate_utf8(const char *in)
+CALLER_OWN char *owl_validate_utf8(const char *in)
 {
   char *out;
   if (g_utf8_validate(in, -1, NULL)) {
-    out = owl_strdup(in);
+    out = g_strdup(in);
   } else {
-    out = owl_strdup("");
+    out = g_strdup("");
   }
   return out;
 }
@@ -732,7 +640,8 @@ int owl_util_can_break_after(gunichar c)
   return 0;
 }
 
-char *owl_escape_highbit(const char *str)
+/* caller must free the return */
+CALLER_OWN char *owl_escape_highbit(const char *str)
 {
   GString *out = g_string_new("");
   unsigned char c;
@@ -760,7 +669,7 @@ static int owl_getline_internal(char **s, FILE *fp, int newline)
     c = getc(fp);
     if ((target + 1) > size) {
       size += BUFSIZ;
-      *s = owl_realloc(*s, size);
+      *s = g_renew(char, *s, size);
     }
     if (c == EOF)
       break;
@@ -777,7 +686,7 @@ static int owl_getline_internal(char **s, FILE *fp, int newline)
 
 /* Read a line from fp, allocating memory to hold it, returning the number of
  * byte read.  *s should either be NULL or a pointer to memory allocated with
- * owl_malloc; it will be owl_realloc'd as appropriate.  The caller must
+ * g_malloc; it will be g_renew'd as appropriate.  The caller must
  * eventually free it.  (This is roughly the interface of getline in the gnu
  * libc).
  *
@@ -795,7 +704,7 @@ int owl_getline_chomp(char **s, FILE *fp)
 }
 
 /* Read the rest of the input available in fp into a string. */
-char *owl_slurp(FILE *fp)
+CALLER_OWN char *owl_slurp(FILE *fp)
 {
   char *buf = NULL;
   char *p;
@@ -803,7 +712,7 @@ char *owl_slurp(FILE *fp)
   int count;
 
   while (1) {
-    buf = owl_realloc(buf, size + BUFSIZ);
+    buf = g_renew(char, buf, size + BUFSIZ);
     p = &buf[size];
     size += BUFSIZ;
 
@@ -815,6 +724,21 @@ char *owl_slurp(FILE *fp)
   return buf;
 }
 
+int owl_util_get_colorpairs(void) {
+#ifndef NCURSES_EXT_COLORS
+  /* Without ext-color support (an ABI change), ncurses only supports 256
+   * different color pairs. However, it gives us a larger number even if your
+   * ncurses is compiled without ext-color. */
+  return MIN(COLOR_PAIRS, 256);
+#else
+  /* In theory ext-color support should allow for 256 color pairs, but there's
+   * a bug in ncurses. The fix is now upstreamed. The commit which introduced
+   * this change (and simultaneously reverted the wbkgrndset change to fix the
+   * build on Solaris) should be reverted when we ship a new locker ncurses. */
+  return MIN(COLOR_PAIRS, 256);
+#endif
+}
+
 gulong owl_dirty_window_on_signal(owl_window *w, gpointer sender, const gchar *detailed_signal)
 {
   return owl_signal_connect_object(sender, detailed_signal, G_CALLBACK(owl_window_dirty), w, G_CONNECT_SWAPPED);
